use std::alloc::{alloc, dealloc, Layout};

extern crate wee_alloc;

extern {
    fn fill_string(p: *mut u8);
    fn log_num_idxs(i: usize);
}


#[global_allocator]
static ALLOC: wee_alloc::WeeAlloc = wee_alloc::WeeAlloc::INIT;

const LIST_END: u16 = 65535;

fn unwrap_or_abort<T>(v: Option<T>) -> T {
    match v {
        Some(v) => v,
        None => std::process::abort(),
    }
}

unsafe fn alloc_ary<T>(sz: usize) -> &'static mut [T] {
    let layout = unwrap_or_abort(Layout::from_size_align(
            sz*std::mem::size_of::<T>(),
            std::mem::align_of::<T>()
        ).ok());
    let ptr = alloc(layout) as *mut T;
    std::slice::from_raw_parts_mut(ptr, sz)
}

unsafe fn dealloc_ary<'a, T>(v: &'a mut [T]) {
    let layout = unwrap_or_abort(Layout::from_size_align(
            v.len()*std::mem::size_of::<T>(),
            std::mem::align_of::<T>()
        ).ok());
    dealloc(v.as_ptr() as *mut u8, layout);
}

// NOTE: you must use this like a = realloc_ary(a, 100)
// otherwise you will suffer from use after free
unsafe fn realloc_ary<'a, T>(v: &'a mut [T], sz: usize) -> &'static mut [T] {
    if v.len() == sz {
        unsafe { std::slice::from_raw_parts_mut(v.as_mut_ptr(), v.len()) }
    } else {
        if v.len() > 0 {
            unsafe { dealloc_ary(v) };
        }
        unsafe { alloc_ary(sz) }
    }
}

// webassembly version of all the wordle logic
// because the javascript version is slow as hell on firefox

// TODO wrap all this in a struct to reduce the insane amount of
// unsafe in the code
pub static mut LOOKUP: &'static mut [u8] = &mut [];
pub static mut STRINGS: &'static mut [u8] = &mut [];
pub static mut STR_IDXS: &'static mut [*const u8] = &mut [];
pub static mut VALID_STRS: &'static mut [u16] = &mut [];

// TODO maybe fiddle with the lifetimes eventually
pub struct Solver {
    lookup: &'static mut [u8],
    strings: &'static mut [u8],
    idxs: &'static mut [*const u8],
    valid_words: &'static mut [u16],
}

impl Solver {
    fn new() -> Solver {
        Solver {
            strings: &mut [],
            idxs: &mut [],

            valid_words: &mut [],
            lookup: &mut [],
        }
    }

    // NOTE: calls fill_string to get the string from Javascript
    fn init(&mut self, str_sz: usize) {
        unsafe {
            self.strings = realloc_ary(self.strings, str_sz);
            fill_string(self.strings.as_mut_ptr());
            self.init_index();
            log_num_idxs(self.idxs.len());
            self.reset();
        }
    }

    fn init_index(&mut self) {
        let mut last_alpha = false;
        let mut num_words = 0;
        for v in self.strings.iter() {
            let cur_alpha = *v >= ('a' as u8) && *v <= ('z' as u8);
            if cur_alpha && !last_alpha {
                num_words += 1;
            }
            last_alpha = cur_alpha;
        }

        unsafe {
            self.idxs = realloc_ary(self.idxs, num_words);
        }

        let mut idx = 0;
        for (i, v) in self.strings.iter().enumerate() {
            let cur_alpha = *v >= ('a' as u8) && *v <= ('z' as u8);
            if cur_alpha && !last_alpha {
                self.idxs[idx] = &self.strings[i] as *const u8;
                idx += 1;
            }
            last_alpha = cur_alpha;
        }
    }

    fn reset(&mut self) {
        unsafe {
            self.valid_words = realloc_ary(self.valid_words, self.idxs.len());
        }
        for (i, v) in self.valid_words.iter_mut().enumerate() {
            *v = i as u16;
        }
    }

    fn count_valid_words(&self) -> usize {
        self.valid_words.iter().take_while(|&v| *v != LIST_END).count()
    }

    fn eliminate_words(&mut self, guess_idx: usize, guess_result: u8) {
        let num_valid_words = self.count_valid_words();
        let mut idx = 0;
        for i in 0..num_valid_words {
            if self.lookup[guess_idx * self.idxs.len() + i] == guess_result {
                self.valid_words[idx] = self.valid_words[i];
            }
        }
        if idx != self.valid_words.len() {
            self.valid_words[idx] = LIST_END;
        }
    }
}

#[no_mangle]
pub fn idxs_offset(s: &Solver) -> *const *const u8 {
    s.idxs.as_ptr()
}

#[no_mangle]
pub extern fn init(str_sz: usize) -> *mut Solver {
    let s = Box::new(Solver::new());
    s.init(str_sz);
    let ret = unsafe { s.as_ptr_mut() };
    std::mem::forget(s);
    ret
}

#[no_mangle]
fn precalc(num_words: usize) {
    let mut i = unsafe { IDX };
    let idxs = unsafe { &*STR_IDXS };
    let lookup = unsafe { &mut *LOOKUP };
    let l = idxs.len();

    for _ in 0..num_words {
        if i >= idxs.len() {
            break;
        }
        let guess = idxs[i];
        for (j, reference) in idxs.iter().enumerate() {
            lookup[i*l + j] = calc_match(guess, *reference);
        }
        i += 1;
    }

    unsafe {
        IDX = i;
    }
}

#[no_mangle]
fn precalc_done() -> bool {
    unsafe {
        IDX >= STR_IDXS.len()
    }
}

#[no_mangle]
fn calc_match(guess: *const u8, reference: *const u8) -> u8 {
    let mut unmatched = vec![0u8; 26];
    let mut ret = 0;
    let mut matched = 0;

    let mut mul = 1;
    for i in 0..5 {
        unsafe {
            if *(guess.offset(i)) == *(reference.offset(i)) {
                ret += 1*mul;
                matched |= 1 << i;
            } else {
                unmatched[*(reference.offset(i)) as usize - 'a' as usize] += 1;
            }
        }
        mul *= 3;
    }

    mul = 1;
    for i in 0..5 {
        if (matched & (1 << i)) == 0 {
            let idx = unsafe { *(guess.offset(i)) as usize - ('a' as usize) };
            if unmatched[idx] > 0 {
                ret += 2*mul;
                unmatched[idx] -= 1;
            }
        }
        mul *= 3;
    }

    ret
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn it_works() {
        let result = add(2, 2);
        assert_eq!(result, 4);
    }
}
